/*
	
	ArpRootLayout.cpp
	
	Copyright (c)1998 by Angry Red Planet.

	This code is distributed under a modified form of the
	Artistic License.  A copy of this license should have
	been included with it; if this wasn't the case, the
	entire package can be obtained at
	<URL:http://www.angryredplanet.com/>.

	The top of a layout tree.
*/

#ifndef ARPLAYOUT_ARPROOTLAYOUT_H
#include "ArpLayout/ArpRootLayout.h"
#endif

#ifndef ARPKERNEL_ARPDEBUG_H
#include "ArpKernel/ArpDebug.h"
#endif

#ifndef ARPKERNEL_ARPCOLOR_H
#include "ArpKernel/ArpColor.h"
#endif

#ifndef _WINDOW_H
#include <be/interface/Window.h>
#endif

#ifndef _MENU_H
#include <be/interface/Menu.h>
#endif

#include <be/support/Autolock.h>
#include <float.h>

ArpMOD();

static const char* GlobalsParam = "ArpRootLayoutable:Globals";

/* ------ ArpRootLayout constructor and destructor ------
 *
 * The various ways to create and destroy ArpRootLayout
 * objects.
 */
 
ArpRootLayout::ArpRootLayout(BRect frame, const char* name,
							 uint32 resizeMask, uint32 flags)
	: ArpLView(frame,name,resizeMask,flags)
{
	ArpD(cdb << ADH << "Creating root layout " << name << endl);
	
	mGlobals.AddRGBColor("StdBackColor",
						ArpColor(ui_color(B_PANEL_BACKGROUND_COLOR)));
	mGlobals.AddRGBColor("StdTextColor",
						ArpColor(0x00,0x00,0x00));
	mGlobals.AddRGBColor("FillBackColor",
						ArpColor(0xff,0xff,0xff));
	mGlobals.AddRGBColor("FillTextColor",
						ArpColor(0x00,0x40,0x40));
	mGlobals.AddFont("BasicFont", be_plain_font);
	mGlobals.AddFont("BoldFont", be_bold_font);
	
	// Standard menu style
	BFont menuFont(be_bold_font);
	ArpColor menuBack(ui_color(B_PANEL_BACKGROUND_COLOR));
	ArpColor menuText(0x00,0x00,0x00);
	struct menu_info minfo;
	if( get_menu_info(&minfo) == B_NO_ERROR ) {
		menuFont.SetFamilyAndStyle(minfo.f_family,minfo.f_style);
		menuFont.SetSize(minfo.font_size);
		menuBack = minfo.background_color;
	}
	mGlobals.AddRGBColor("MenuBackColor", menuBack);
	mGlobals.AddRGBColor("MenuTextColor", menuText);
	mGlobals.AddFont("MenuFont", &menuFont);
	
	ArpD(cdb << ADH << "Placing root globals into layout..." << endl);
	SetGlobals(&mGlobals);
	
	ArpD(cdb << ADH << "Setting layout to current bounds..." << endl);
	SetLayout(Bounds());
}

ArpRootLayout::ArpRootLayout(BMessage* data)
	: ArpLView(data)
{
	if( data ) {
		data->FindMessage(GlobalsParam, &mGlobals);
	}
	
	SetGlobals(&mGlobals);
	SetLayout(Bounds());
}

ArpRootLayout::~ArpRootLayout()
{
	SetLayoutInhibit(true);
	SetGlobals(NULL);
}

/* ------------ ArpRootLayout archiving ------------
 *
 * Archiving and retrieving ArpRootLayout objects.
 */
BArchivable* ArpRootLayout::Instantiate(BMessage* archive)
{
	if ( validate_instantiation(archive, "ArpRootLayout") ) 
		return new ArpRootLayout(archive); 
	return NULL;
}

status_t ArpRootLayout::Archive(BMessage* data, bool deep)
{
	status_t status = ArpLView::Archive(data, deep);
	if( status == B_NO_ERROR )
		status = data->AddMessage(GlobalsParam, &mGlobals);
	if( status == B_NO_ERROR )
		status = data->AddString("class","ArpRootLayout");
	return status;
}

void ArpRootLayout::AttachedToWindow()
{
	ArpD(cdb << ADH << "Root " << Name()
				 << ": this view is attached to window "
				 << Window()->Title() << endl);
}

void ArpRootLayout::AllAttached()
{
	ArpD(cdb << ADH << "Root " << Name()
				 << ": all views are now attached to window "
				 << Window()->Title() << endl);
				
	// When an ArpLayout class is first created, it starts
	// with its LayoutInhibit flag sets: this tells it not
	// to do relayouts in response to changes in its
	// children etc., until the flag is turned off.  There
	// are two reasons for doing this.  The most obvious is
	// that it saves a lot of overhead when constructing the
	// initial object hierarchy, as it's not recomputing the
	// layout every time an object is added.  However, there
	// is an even more important reason: many Be objects
	// don't supply sane GetPreferredSize() dimensions until
	// they are added to a window.  By waiting until now
	// to allow the layout to happen, we don't have to worry
	// about cacheing any incorrect dimension information
	// picked up from before the UI was added to the window.
	SetLayoutInhibit(false);
	
	// If this is the top-level view in the window, set up
	// the size limits of the window from the layout.
	if( Parent() == 0 ) SetWindowLimits();
}

void ArpRootLayout::DetachedFromWindow()
{
	ArpD(cdb << ADH << "Root " << Name()
				 << ": all views are now detached from window." << endl);
				
	// Similarily, disable updates when removed from a window.
	SetLayoutInhibit(true);
}

void ArpRootLayout::FrameResized(float width, float height)
{
	if( Window() ) {
		ArpD(cdb << ADH <<"-------------------------------------------" << endl);
		ArpD(cdb << ADH << "ArpRootLayout::FrameResized()" << endl);
		ArpD(cdb << ADH << "  width=" << width << ", height=" << height << endl);
		BRect bounds = Bounds();
		ArpD(cdb << ADH << "  Resizing to bounds: " << bounds << endl);
		Window()->BeginViewTransaction();
		SetLayout(bounds);
		Window()->EndViewTransaction();
	}
}

void ArpRootLayout::WindowActivated(bool state)
{
	BView::WindowActivated(state);
	inherited::SetLayoutActivated(state);
}

void ArpRootLayout::MessageReceived(BMessage *message)
{
	if( message ) {
		switch(message->what) {
			case ARP_GLOBALS_MSG: {
				ArpMessage tmp = *message;
				UpdateGlobals(tmp);
			} break;
			default: {
				BView::MessageReceived(message);
			} break;
		}
	}
}

void ArpRootLayout::Draw(BRect updateRect)
{
	ArpD(cdb << ADH << "Root " << Name()
				 << ": drawing window with frame "
				 << updateRect
				 << " (bounds=" << Bounds() << ")" << endl);
}

void ArpRootLayout::FrameMoved(BPoint new_position)
{
	ArpD(cdb << ADH << "Root " << Name()
				 << ": frame moved to "
				 << new_position << endl);
}

void ArpRootLayout::SetWindowLimits()
{
	if( Window() ) {
		float minw=0,minh=0,maxw=0,maxh=0;
		ArpLayoutDimens dimens = LayoutDimens();
		Window()->GetSizeLimits(&minw,&maxw,&minh,&maxh);
		Window()->SetSizeLimits(dimens.min_width,maxw,dimens.min_height,maxh);
		Window()->SetZoomLimits(dimens.pref_width, dimens.pref_height);
		BRect bounds = Window()->Bounds();
		ArpD(cdb << ADH << "SetWindowLimits():" << endl);
		ArpD(cdb << ADH << "  Bounds: " << bounds << ", minw=" << minw
						<< ", minh=" << minh << ", maxw=" << maxw
						<< ", maxh=" << maxh << endl);
		ArpD(cdb << ADH << "  Layout minw=" << dimens.min_width
						<< ", minh=" << dimens.min_height << endl);
		float bw = bounds.Width();
		float bh = bounds.Height();
		if( bw < dimens.min_width || bh < dimens.min_height ) {
			if( bw < dimens.min_width ) bw = dimens.min_width;
			if( bh < dimens.min_height ) bh = dimens.min_height;
			Window()->ResizeTo(bw, bh);
		}
	}
}

status_t ArpRootLayout::AddGlobals(const ArpMessage* gl)
{
	status_t res = B_NO_ERROR;
	if( gl ) {
		ArpD(cdb << ADH << "Adding global values: " << *gl << endl);
		char* name;
		ulong type;
		long count;
		for( int32 i=0; !gl->GetInfo(B_ANY_TYPE,i,&name,&type,&count); i++ ) {
			ArpD(cdb << ADH << "Adding global " << name
							<< " (type=" << type <<")." << endl);
			if( !mGlobals.HasData(name,type) ) {
				const void* data;
				ssize_t size;
				if( !(res=gl->FindData(name,type,&data,&size)) ) {
					res = mGlobals.AddData(name,type,data,size);
				} else {
					ArpD(cdb << ADH << "Ehhh?  Error getting global: res="
								<< res << "." << endl);
				}
			} else {
				type = 0;
				int32 count;
				mGlobals.GetInfo(name,&type,&count);
				ArpD(cdb << ADH << "Whoops, already have it!  Type is "
							<< type << "." << endl);
			}
		}
	}
	
	return res;
}

void ArpRootLayout::UpdateGlobals(const ArpMessage& newVals)
{
	bool oldInhibit = LayoutInhibit();
	SetLayoutInhibit(true);
	
	mGlobals.Update(newVals);
	if( Window() ) Window()->BeginViewTransaction();
	
	ArpD(cdb << ADH << "Global update msg = " << newVals
				<< endl);
	ArpD(cdb << ADH << "Final global values = " << mGlobals
				<< endl);
	ArpD(cdb << ADH << "-------------------------------------" << endl);
	ArpD(cdb << ADH << "Refreshing globals:" << endl);
	RefreshGlobals(&newVals);
	
	SetLayoutInhibit(oldInhibit);
	
	if( Window() ) {
		ArpD(cdb << ADH << "-------------------------------------" << endl);
		ArpD(cdb << ADH << "Getting dimensions:" << endl);
		BRect bounds = Bounds();
		ArpLayoutDimens dimens = LayoutDimens();
		if( Parent() == 0 ) {		// top-level view in window?
			if( bounds.Width() < dimens.min_width ||
				bounds.Height() < dimens.min_height ) {
				ArpD(cdb << ADH << "-------------------------------------"
						<< endl);
				ArpD(cdb << ADH << "Resizing window:" << endl);
				Window()->ResizeTo(dimens.min_width,dimens.min_height);
				bounds = Window()->Bounds();
			}
			SetWindowLimits();
		}
		ArpD(cdb << ADH << "-------------------------------------" << endl);
		ArpD(cdb << ADH << "Setting layout:" << endl);
		SetLayout(bounds);
		ArpD(cdb << ADH << "-------------------------------------" << endl);
		ArpD(cdb << ADH << "Enabling updates:" << endl);
		Window()->EndViewTransaction();
		ArpD(cdb << ADH << "-------------------------------------" << endl);
		ArpD(cdb << ADH << "Done!" << endl);
	}
}
