/*
	
	ArpScrollArea.c
	
	Copyright (c)1998 by Angry Red Planet.

	This code is distributed under a modified form of the
	Artistic License.  A copy of this license should have
	been included with it; if this wasn't the case, the
	entire package can be obtained at
	<URL:http://www.angryredplanet.com/>.

	A layout class that places scroll bars around an interior
	area.  Only one child is allowed, and it must possess its
	own BView to which the scrollers can be attached.
*/

#ifndef ARPLAYOUT_ARPSCROLLAREA_H
#include <ArpLayout/ArpScrollArea.h>
#endif

#ifndef ARPKERNEL_ARPDEBUG_H
#include <ArpKernel/ArpDebug.h>
#endif

#ifndef _AUTOLOCK_H
#include <be/support/Autolock.h>
#endif

#ifndef _SCROLLBAR_H
#include <be/interface/ScrollBar.h>
#endif

#include <algobase.h>
#include <string.h>

ArpMOD();

const char* ArpScrollArea::ScrollHorizontalP = "ScrollHorizontal";
const char* ArpScrollArea::ScrollVerticalP = "ScrollVertical";
const char* ArpScrollArea::InsetCornerP = "InsetCorner";
const char* ArpScrollArea::BorderStyleP = "BorderStyle";

//static const char* VScrollParam = "ArpScrollArea:VScroll";
//static const char* HScrollParam = "ArpScrollArea:HScroll";

/* ------ ArpScrollArea constructor and destructor ------
 *
 * The various ways to create and destroy ArpScrollArea
 * objects.
 */
 
ArpScrollArea::ArpScrollArea(const char* name)
		: ArpLayoutable(name)
{
	PG_ScrollHorizontal = PG_ScrollVertical = PG_InsetCorner
		= PG_BorderStyle = (const char*)0;
	PV_ScrollHorizontal = PV_ScrollVertical = PV_InsetCorner = false;
	PV_BorderStyle = B_FANCY_BORDER;

	scroll_h = new BScrollBar(BRect(0,0,100,B_H_SCROLL_BAR_HEIGHT),
							"hscroll", NULL, 0, 0, B_HORIZONTAL);
	scroll_v = new BScrollBar(BRect(0,0,B_V_SCROLL_BAR_WIDTH,100),
							"vscroll", NULL, 0, 0, B_VERTICAL);
	if( scroll_h ) {
		scroll_h->Hide();
		scroll_h->SetResizingMode(B_FOLLOW_NONE);
	}
	if( scroll_v ) {
		scroll_v->Hide();
		scroll_v->SetResizingMode(B_FOLLOW_NONE);
	}
	do_vert = do_horiz = inset_corner = false;
	border = 0;
	s_width = s_height = 0;
}

ArpScrollArea::ArpScrollArea(BMessage* data)
	: ArpLayoutable(data)
{
	/* XXX We need to initialize these BEFORE!!
	PG_ScrollHorizontal = PG_ScrollVertical = PG_InsetCorner
		= PG_BorderStyle = NULL;
	PV_ScrollHorizontal = PV_ScrollVertical = PV_InsetCorner = false;
	PV_BorderStyle = B_FANCY_BORDER;
	*/
	
	scroll_h = new BScrollBar(BRect(0,0,1,B_H_SCROLL_BAR_HEIGHT),
							"hscroll", NULL, 0, 0, B_HORIZONTAL);
	scroll_v = new BScrollBar(BRect(0,0,B_V_SCROLL_BAR_WIDTH,1),
							"vscroll", NULL, 0, 0, B_VERTICAL);
	if( scroll_h ) {
		scroll_h->Hide();
		scroll_h->SetResizingMode(B_FOLLOW_NONE);
	}
	if( scroll_v ) {
		scroll_v->Hide();
		scroll_v->SetResizingMode(B_FOLLOW_NONE);
	}
	do_vert = do_horiz = inset_corner = false;
	border = 0;
	s_width = s_height = 0;
}

/* ------------ ArpScrollArea archiving ------------
 *
 * Archiving and retrieving ArpLayoutable objects.
 */
ArpScrollArea* ArpScrollArea::Instantiate(BMessage* archive)
{
	if ( validate_instantiation(archive, "ArpScrollArea") ) 
		return new ArpScrollArea(archive); 
	return NULL;
}

status_t ArpScrollArea::Archive(BMessage* data, bool deep)
{
	status_t status = inherited::Archive(data,deep);
	if( status == B_NO_ERROR )
		status = data->AddString("class","ArpScrollArea");
	return status;
}

BView* ArpScrollArea::get_child(void)
{
	ArpLayoutable* childl = LayoutChildAt(0);
	if( childl ) return childl->OwnerView();
	return NULL;
}

void ArpScrollArea::DoSetParams(const ArpMessage* g,
								const ArpMessage* p)
{
	inherited::DoSetParams(g,p);
	
	ArpD(cdb << ADH << "ArpScrollArea::DoSetParams(), with:" << endl);
	ArpD(if( g ) cdb << ADH << "Globals = " << *g << endl);
	ArpD(if( p ) cdb << ADH << "Params = " << *p << endl);
	
	bool scr_changed = false;
	status_t res;
	res = ExtractParam(p,ScrollHorizontalP,B_BOOL_TYPE,
						g,&PG_ScrollHorizontal,&PV_ScrollHorizontal);
	if( !res ) scr_changed = true;
	res = ExtractParam(p,ScrollVerticalP,B_BOOL_TYPE,
						g,&PG_ScrollVertical,&PV_ScrollVertical);
	if( !res ) scr_changed = true;
	if( scr_changed ) update_scroll_bars();
	
	res = ExtractParam(p,InsetCornerP,B_BOOL_TYPE,
						g,&PG_InsetCorner,&PV_InsetCorner);
	if( !res ) InvalidateDimens();
	
	res = ExtractParam(p,BorderStyleP,B_INT32_TYPE,
						g,&PG_BorderStyle,&PV_BorderStyle);
	if( !res ) InvalidateDimens();
}

bool ArpScrollArea::DoParams(ArpMessage& p, const char* name)
{
	bool res = inherited::DoParams(p,name);
	if( !name || !strcmp(name,ScrollHorizontalP) ) {
		if( p.AddBool(ScrollHorizontalP,PV_ScrollHorizontal) ) {
			res = false;
		}
		if( name ) return res;
	}
	if( !name || !strcmp(name,ScrollVerticalP) ) {
		if( p.AddBool(ScrollVerticalP,PV_ScrollVertical) ) {
			res = false;
		}
		if( name ) return res;
	}
	if( !name || !strcmp(name,InsetCornerP) ) {
		if( p.AddBool(InsetCornerP,PV_InsetCorner) ) {
			res = false;
		}
		if( name ) return res;
	}
	if( !name || !strcmp(name,BorderStyleP) ) {
		if( p.AddInt32(BorderStyleP,PV_BorderStyle) ) {
			res = false;
		}
		if( name ) return res;
	}
	return res;
}

void ArpScrollArea::update_scroll_bars(void)
{
	BView* childv = get_child();
	bool changed = false;
	if( PV_ScrollHorizontal && scroll_h && scroll_h->IsHidden() ) {
		ArpD(cdb << ADH << "ArpScrollArea: Showing hscroll" << endl);
		if( childv ) scroll_h->SetTarget(childv);
		scroll_h->Show();
		changed = true;
	} else if( !PV_ScrollHorizontal && scroll_h && !scroll_h->IsHidden() ) {
		ArpD(cdb << ADH << "ArpScrollArea: Hiding hscroll" << endl);
		scroll_h->SetTarget((BView*)NULL);
		scroll_h->Hide();
		changed = true;
	}
	if( PV_ScrollVertical && scroll_v && scroll_v->IsHidden() ) {
		ArpD(cdb << ADH << "ArpScrollArea: Showing vscroll" << endl);
		if( childv ) scroll_v->SetTarget(childv);
		scroll_v->Show();
		changed = true;
	} else if( !PV_ScrollVertical && scroll_v && !scroll_v->IsHidden() ) {
		ArpD(cdb << ADH << "ArpScrollArea: Hiding vscroll" << endl);
		scroll_v->SetTarget((BView*)NULL);
		scroll_v->Hide();
		changed = true;
	}
	if( changed ) InvalidateDimens();
}

BScrollBar* ArpScrollArea::HScrollBar(void)
{
	if(this && PV_ScrollHorizontal) return scroll_h;
	return NULL;
}

BScrollBar* ArpScrollArea::VScrollBar(void)
{
	if(this && PV_ScrollVertical) return scroll_v;
	return NULL;
}

ArpLayoutable* ArpScrollArea::AddLayoutChild(ArpLayoutable* v,
						const BMessage& c, ArpLayoutable* before)
{
	assert(CountLayoutChildren() == 0);
	ArpLayoutable* ret = inherited::AddLayoutChild(v,c,before);
	if( v ) {
		assert(v->OwnerView() != NULL);
		if( PV_ScrollVertical && scroll_v )
			scroll_v->SetTarget(v->OwnerView());
		if( PV_ScrollHorizontal && scroll_h )
			scroll_h->SetTarget(v->OwnerView());
	}
	return ret;
}

bool ArpScrollArea::RemoveLayoutChild(ArpLayoutable* child)
{
	bool ret = inherited::RemoveLayoutChild(child);
	if( ret && child ) {
		if( scroll_v ) scroll_v->SetTarget((BView*)NULL);
		if( scroll_h ) scroll_h->SetTarget((BView*)NULL);
	}
	return ret;
}

void ArpScrollArea::SetFocusShown(bool state, bool andParent)
{
	bool old = FocusShown();
	inherited::SetFocusShown(state, false);
	if( InView() && old != state ) DrawLayout(InView(), InView()->Bounds());
}


void ArpScrollArea::SetLayoutActivated(bool state)
{
	bool old = LayoutActivated();
	inherited::SetLayoutActivated(state);
	if( InView() && old != state && FocusShown() ) {
		DrawLayout(InView(), InView()->Bounds());
	}
}

void ArpScrollArea::AttachView(BView* par_view, BView* before)
{
	ArpD(cdb << ADH << "*** ArpScrollArea::AttachView()" << endl);
	inherited::AttachView(par_view, before);
	if( par_view ) {
		if( scroll_h ) par_view->AddChild(scroll_h);
		if( scroll_v ) par_view->AddChild(scroll_v);
	} else if( InView() ) {
		if( scroll_h ) InView()->RemoveChild(scroll_h);
		if( scroll_v ) InView()->RemoveChild(scroll_v);
	}
}

void ArpScrollArea::ComputeDimens(void)
{
	inherited::ComputeDimens();
	ArpD(cdb << ADH << "scr min_width = " << cur_dimens.min_width
			<< ", min_height = " << cur_dimens.min_height << endl);

	float dimw=0, dimh=0;
	if( (PV_ScrollHorizontal || PV_InsetCorner) && scroll_h )
		scroll_h->GetPreferredSize(&dimw,&dimh);
	cur_dimens.min_height += dimh+1;
	cur_dimens.pref_height += dimh+1;
	cur_dimens.min_width = max(cur_dimens.min_width,(dimh*3)+4);
	cur_dimens.pref_width = max(cur_dimens.pref_width,(dimh*3)+4);
	s_height = dimh;
	
	ArpD(cdb << ADH << "Computed hscroll dimens:" << endl
				<< "  dimw=" << dimw << " dimh=" << dimh
				<< " min_w=" << cur_dimens.min_width
				<< " min_h=" << cur_dimens.min_height << endl);
	
	dimw = dimh = 0;
	if( (PV_ScrollVertical || PV_InsetCorner) && scroll_v )
		scroll_v->GetPreferredSize(&dimw,&dimh);
	cur_dimens.min_width += dimw+1;
	cur_dimens.pref_width += dimw+1;
	cur_dimens.min_height = max(cur_dimens.min_height,(dimw*3)+4);
	cur_dimens.pref_height = max(cur_dimens.pref_height,(dimw*3)+4);
	s_width = dimw;

	ArpD(cdb << ADH << "Computed vscroll dimens:" << endl
				<< "  dimw=" << dimw << " dimh=" << dimh
				<< " min_w=" << cur_dimens.min_width
				<< " min_h=" << cur_dimens.min_height << endl);
	
	if( (PV_ScrollHorizontal && PV_ScrollVertical)
		|| (PV_InsetCorner && (PV_ScrollHorizontal || PV_ScrollVertical)) ) {
		cur_dimens.min_width = max(cur_dimens.min_width,
									s_height + s_width + 4);
		cur_dimens.min_height = max(cur_dimens.min_height,
									s_height + s_width + 4);
		cur_dimens.pref_width = max(cur_dimens.pref_width,
									s_height + s_width + 4);
		cur_dimens.pref_height = max(cur_dimens.pref_height,
									s_height + s_width + 4);
	}
	
	if( PV_BorderStyle == B_FANCY_BORDER ) {
		cur_dimens.min_width += 4;
		cur_dimens.min_height += 4;
		cur_dimens.pref_width += 4;
		cur_dimens.pref_height += 4;
		cur_dimens.max_width += 4;
		cur_dimens.max_height += 4;
	} else if( PV_BorderStyle == B_PLAIN_BORDER ) {
		cur_dimens.min_width += 2;
		cur_dimens.min_height += 2;
		cur_dimens.pref_width += 2;
		cur_dimens.pref_height += 2;
		cur_dimens.max_width += 2;
		cur_dimens.max_height += 2;
	}
	
	ArpD(cdb << ADH << "final min_width = " << cur_dimens.min_width
			<< ", min_height = " << cur_dimens.min_height << endl);
	ArpD(cdb << ADH << "final pref_width = " << cur_dimens.pref_width
			<< ", pref_height = " << cur_dimens.pref_height << endl);
}

void ArpScrollArea::Layout(void)
{
	if( PV_BorderStyle != B_NO_BORDER && InView() ) {
		// If we are drawing a frame around the children, we need to
		// invalidate our containing view to be sure the frame is
		// completely redrawn.  Curiously, calling Invalidate() with
		// an explicit rectangle seems to cause EVERYTHING to be redrawn...
		// including all the parents!!!  So it's better to just invalidate
		// the whole view, it seems.  *sigh*
		ArpD(cdb << ADH << "Invalidating parent view: " << InView()->Name()
				<< " (" << InView() << ")" << endl);
		if( mLastFrame.IsValid() ) {
			ArpD(cdb << ADH << "Previous rectangle: " << mLastFrame << endl);
			//InView()->Invalidate(mLastFrame);
		}
		mLastFrame = LayoutFrame();
		ArpD(cdb << ADH << "New rectangle: " << mLastFrame << endl);
		//InView()->Invalidate(mLastFrame);
		InView()->Invalidate();
	}
	
	BRect frm = LayoutBounds();
	int32 num = CountLayoutChildren();

	ArpD(cdb << ADH << "Frame=" << frm << " s_width=" << s_width
				<< " s_height=" << s_height << endl);
	
	if( PV_BorderStyle == B_FANCY_BORDER ) {
		frm.left += 2;
		frm.top += 2;
		frm.right -= 2;
		frm.bottom -= 2;
	} else if( PV_BorderStyle == B_PLAIN_BORDER ) {
		frm.left += 1;
		frm.top += 1;
		frm.right -= 1;
		frm.bottom -= 1;
	}
	
	if( PV_ScrollHorizontal && scroll_h ) frm.bottom -= s_height;
	if( PV_ScrollVertical && scroll_v ) frm.right -= s_width;

	if( PV_ScrollHorizontal && scroll_h && frm.Width() > 0 ) {
		ArpD(cdb << ADH << "Moving hscroll to (" << frm.right << ","
					<< frm.top << ") size=(" << s_width
					<< "," << frm.Height() << ")" << endl);
		ArpD(cdb << ADH << "Hscroll parent="
					<< (void*)scroll_h->Parent()
					<< " ishidden=" << scroll_h->IsHidden() << endl);
		scroll_h->MoveTo(frm.left,frm.bottom+1);
		float inset = (PV_InsetCorner && !PV_ScrollVertical) ? s_width : 0;
		float width = frm.Width()-inset;
		if( width < s_height+4 ) width = s_height+4;
		scroll_h->ResizeTo(width,s_height);
	}
	if( PV_ScrollVertical && scroll_v && frm.Height() > 0 ) {
		ArpD(cdb << ADH << "Moving vscroll to (" << frm.left << ","
					<< frm.bottom << ") size=(" << frm.Width()
					<< "," << s_height << ")" << endl);
		ArpD(cdb << ADH << "Vscroll parent="
					<< (void*)scroll_v->Parent()
					<< " ishidden=" << scroll_v->IsHidden() << endl);
		scroll_v->MoveTo(frm.right+1,frm.top);
		float inset = (PV_InsetCorner && !PV_ScrollHorizontal) ? s_height : 0;
		float height = frm.Height()-inset;
		if( height < s_width+4 ) height = s_width+4;
		scroll_v->ResizeTo(s_width,height);
	}
	
	if( num >= 1 ) {
		ArpLayoutable* child = LayoutChildAt(0);
		ArpD(cdb << ADH << "Moving child to " << frm << endl);
		if( child ) child->SetLayout(frm);
	}
	
	ArpD(cdb << ADH << "cframe=" << (get_child() ? get_child()->Frame():BRect())
				<< " vframe=" << (scroll_v ? scroll_v->Frame():BRect())
				<< " hframe=" << (scroll_h ? scroll_h->Frame():BRect())
				<< endl);
}

void ArpScrollArea::DrawLayout(BView* inside, BRect region)
{
	if( !inside ) return;
	inside->PushState();
	inside->SetPenSize(1.0);
	BRect rect(LayoutBounds());
	ArpD(cdb << ADH << "Drawing frame in bounds: " << rect << endl);
	bool show_focus = LayoutActivated() ? FocusShown() : false;
	
	if( PV_BorderStyle == B_FANCY_BORDER ) {
		if( show_focus ) {
			inside->SetHighColor(ui_color(B_KEYBOARD_NAVIGATION_COLOR));
			inside->StrokeRect(rect);
		} else {
			inside->SetHighColor(tint_color(ui_color(B_PANEL_BACKGROUND_COLOR),
											B_LIGHTEN_MAX_TINT));
			inside->StrokeLine(BPoint(rect.right,rect.top),
							   BPoint(rect.right,rect.bottom));
			inside->StrokeLine(BPoint(rect.left,rect.bottom),
							   BPoint(rect.right,rect.bottom));
			inside->SetHighColor(tint_color(ui_color(B_PANEL_BACKGROUND_COLOR),
											B_DARKEN_1_TINT));
			inside->StrokeLine(BPoint(rect.left,rect.top),
							   BPoint(rect.left,rect.bottom));
			inside->StrokeLine(BPoint(rect.left,rect.top),
							   BPoint(rect.right,rect.top));
		}
		inside->SetHighColor(tint_color(ui_color(B_PANEL_BACKGROUND_COLOR),
										B_DARKEN_2_TINT));
		rect.left += 1;
		rect.top += 1;
		rect.right -= 1;
		rect.bottom -= 1;
		inside->StrokeRect(rect);
	} else if( PV_BorderStyle == B_PLAIN_BORDER ) {
		if( show_focus ) {
			inside->SetHighColor(ui_color(B_KEYBOARD_NAVIGATION_COLOR));
		} else {
			inside->SetHighColor(tint_color(ui_color(B_PANEL_BACKGROUND_COLOR),
											B_DARKEN_2_TINT));
		}
		inside->StrokeRect(rect);
	}
	inside->PopState();
}
